import numpy as np


from . import fluxes


def transmissive_outflow(surface_flux=fluxes.lax_friedrichs_flux):
    """
    Create a transmissive boundary by setting left and right ("ghost") states the same
    and computing the corresponding surface flux.

    For details see Section 9.2.5 of the book:
    - Eleuterio F. Toro (2001)
      Shock-Capturing Methods for Free-Surface Shallow Flows
      1st edition
      ISBN 0471987662

    :param surface_flux: Function used to compute numerical flux between adjacent cell states.
    """
    def _flux_function(u_l, u_r, xx, t, f, f_prime):
        """

        :param u_l:
        :param u_r:
        :param xx:
        :param t:
        :param f:
        :param f_prime:
        :return:
        """
        return surface_flux(u_l, u_l, xx, t, f, f_prime)

    return _flux_function

def reflective_outflow(surface_flux=fluxes.lax_friedrichs_flux):
    """
    Create a transmissive boundary by reflecting the velocity between left and right ("ghost")
   and computing the corresponding surface flux.

    For details see Section 9.2.5 of the book:
    - Eleuterio F. Toro (2001)
      Shock-Capturing Methods for Free-Surface Shallow Flows
      1st edition
      ISBN 0471987662

    :param surface_flux: Function used to compute numerical flux between adjacent cell states.
    """
    def _flux_function(u_l, u_r, xx, t, f, f_prime):
        h_l, q_l = u_l

        return surface_flux(u_l, np.array((h_l, -q_l)), xx, t, f, f_prime)

    return _flux_function



import numpy as np


def lax_friedrichs_flux(u_l, u_r, xx, t, f, f_prime):
    """
    See http://www.clawpack.org/riemann_book/html/Approximate_solvers.html#Lax-Friedrichs-(LF)-and-local-Lax-Friedrichs-(LLF)

    :param u_l: Left state
    :param u_r: Right state
    :param xx: Domain
    :param t: Current time
    :param f: Hyperbolic flux function
    :param f_prime: Jacobian of hyperbolic flux function
    :return:
    """
    eig_left = np.abs(np.linalg.eigvalsh(f_prime(u_l, xx, t))).max()
    eig_right = np.abs(np.linalg.eigvalsh(f_prime(u_r, xx, t))).max()

    alpha = max(eig_left, eig_right)

    return 0.5 * (f(u_r, xx, t) + f(u_l, xx, t) + alpha * (u_l - u_r))


import binascii
import csv
import os
import time


import numpy as np


from . import bc
from . import fluxes
from . import quadrature
from . import rk
from . import swe_1d


class SWEFlowRunner:
    """
    Helper class for running several simulations in sequence or parallel.
    """
    def __init__(
            self,
            b,
            b_x,
            initial_height,
            xlims,
            tlims,
            gravity=9.81,
            inflow_amplitude=0.05,
            inflow_smoothness=1.0,
            inflow_peak_time=1.0,
            save_dir='./out',
    ) -> None:
        self.gravity = gravity
        self.inflow_amplitude = inflow_amplitude
        self.inflow_smoothness = inflow_smoothness
        self.inflow_peak_time = inflow_peak_time
        self.b = b
        self.b_x = b_x
        self.initial_height = initial_height
        self.xlims = xlims
        self.tlims = tlims

        # Used to save files in unique folders organized by time. Folder name format is
        # "<epoch timestamp>_<six-digit random hex>"
        #
        self._save_dir = os.path.join(
            save_dir,
            '{}_{}'.format(int(time.time()), binascii.b2a_hex(os.urandom(4)).decode())
        )

        assert(inflow_amplitude > 0)

    def q_bc(self, t):
        """
        Describes a Gaussian inflow, where the function transitions to a constant value upon
        attaining its maximum value.

        :param t: Time
        :return:
        """
        tt = np.array(t)
        return self.inflow_amplitude * np.exp(
            -np.square(
                np.minimum(tt, self.inflow_peak_time * np.ones(tt.shape)) - self.inflow_peak_time
            ) / (2 * np.square(self.inflow_smoothness))
        )

    def initial_condition(self, x):
        """
        Creates initial conditions for (h, uh).

        :param x: Computational domain
        :return:
        """
        # Horizontal water surface
        #
        initial_height = self.initial_height * np.ones(x.shape) - self.b(x)

        # Start with whatever flow is prescribed by our inflow BC
        #
        initial_flow = self.q_bc(0) * np.ones(x.shape)

        ic = np.array((
            initial_height,
            initial_flow,
        ))

        # Verify consistency of initial condition
        #
        if not np.allclose(ic[1][0], self.q_bc(0)):
            raise ValueError('Initial flow condition must match prescribed inflow.')

        return ic

    def run(self):
        print('Writing output files to {}'.format(self._save_dir))
        os.makedirs(self._save_dir, exist_ok=True)

        # Write solver parameters to file
        #
        with open(os.path.join(self._save_dir, 'parameters.csv'), 'w', newline='') as csvfile:
            params = {
                'gravity': self.gravity,
                'inflow_amplitude': self.inflow_amplitude,
                'inflow_smoothness': self.inflow_smoothness,
                'inflow_peak_time': self.inflow_peak_time,
                'xl': self.xlims[0],
                'xr': self.xlims[1],
                't0': self.tlims[0],
                'tf': self.tlims[1],
            }
            writer = csv.DictWriter(csvfile, fieldnames=params.keys())
            writer.writeheader()
            writer.writerow(params)

        # Instantiate solver with bathymetry
        #
        solver = swe_1d.ShallowWater1D(
            b=self.b,
            b_x=self.b_x,
            gravity=self.gravity
        )

        t_interval_ms = 20
        dt = t_interval_ms / 1000
        surface_flux = fluxes.lax_friedrichs_flux
        solution = solver.solve(
            tspan=self.tlims,
            xspan=self.xlims,
            cell_count=16,
            polydeg=4,
            initial_condition=self.initial_condition,
            intercell_flux=surface_flux,
            left_boundary_flux=swe_1d.ShallowWater1D.bc_prescribed_inflow(
                self.q_bc,
                gravity=self.gravity,
                surface_flux=surface_flux
            ),
            right_boundary_flux=bc.transmissive_outflow(surface_flux=surface_flux),
            quad_rule=quadrature.gll,
            **{
                'method': rk.SSPRK33,
                't_eval': np.arange(self.tlims[0], self.tlims[1] + dt, dt),
                'max_step': dt,  # max time step for ODE solver
                'rtol': 1.0e-6,
                'atol': 1.0e-6,
            }
        )

        sol, XX = solution  # Uppercase variables indicate matrix; lowercase indicates vector
        xx = XX.flatten()
        uu = sol.y.T.reshape(len(sol.t), solver.solution_vars, XX.shape[0], XX.shape[1])
        bb = self.b(xx)
        bb_x = self.b_x(xx)
        with open(os.path.join(self._save_dir, 'solution.csv'), 'w', newline='') as csvfile:
            params = {
                't': None,
                'x': None,
                'h': None,
                'q': None,
                'b': None,
                'b_x': None,
            }
            writer = csv.DictWriter(csvfile, fieldnames=params.keys())
            writer.writeheader()
            for n, t in enumerate(sol.t):
                params['t'] = t
                hh = uu[n, 0, :, :].flatten()
                qq = uu[n, 1, :, :].flatten()

                for k, x in enumerate(xx):
                    params['x'] = x
                    params['h'] = hh[k]
                    params['q'] = qq[k]
                    params['b'] = bb[k]
                    params['b_x'] = bb_x[k]
                    writer.writerow(params)

        # Plot solution animation
        #
        ani, plt = solver.plot_animation(solution, frame_interval=t_interval_ms)

        # Save animation to file
        #
        movie_name = os.path.join(self._save_dir, 'swe_1d.gif')
        print('Writing movie to {}...'.format(movie_name))

        ani.save(movie_name, progress_callback=lambda i, n: print(
            f'Saving animation frame {i + 1}/{n}'
        ) if i % 50 == 0 else None)
        print('Animation written to {}.'.format(movie_name))



import numpy as np

from . import simulation

# Physical parameters
#
g = 1.0  # gravity

# Domain
#
tspan = (0.0, 10.0)
xspan = (-1, 1)

# Bathymetry parameters
#
b_smoothness = 0.1
b_amplitude = 0.02
b_slope = 0.05
assert(b_smoothness > 0)

# Initial wave height (including bathymetry)
#
H0 = 0.5
assert(H0 > 0)


def swe_bathymetry(x):
    """
    Describes bathymetry with an upslope which is perturbed by a hyperbolic tangent function.
    """
    return b_slope * x + b_amplitude * np.arctan(x/b_smoothness)


def swe_bathymetry_derivative(x):
    """
    Derivative of swe_bathymetry
    """
    return b_slope * np.ones(x.shape) + b_amplitude * b_smoothness/(1 + np.square(b_smoothness*x)) * np.exp(-np.square(b_smoothness*x))


amplitude = np.linspace(0.01, 0.1, num=9)
smoothness = np.linspace(0.5, 2.0, num=16)

for j, a in enumerate(amplitude):
    for k, s in enumerate(smoothness):
        runner = simulation.SWEFlowRunner(
            b=swe_bathymetry,
            b_x=swe_bathymetry_derivative,
            initial_height=H0,
            xlims=xspan,
            tlims=tspan,
            gravity=g,
            inflow_amplitude=a,
            inflow_smoothness=s,
            inflow_peak_time=4.0,
        )
        print('Running simulation {}/{} with a={}, s={}...'.format(
            (j+1)*(k+1),
            len(amplitude) * len(smoothness),
            a,
            s,
        ))
        runner.run()

print('Simulation execution complete.')



import numpy as np
from scipy import optimize

from . import fluxes
from . import hyperbolic_solver_1d


class ShallowWater1D(hyperbolic_solver_1d.Hyperbolic1DSolver):
    solution_vars = 2

    def __init__(self, b=None, b_x=None, gravity=9.81):
        """
        :param b: Bathymetry function. If not provided, flat bathymetry is assumed.
        :param b_x: Derivative of bathymetry function. If not provided, flat bathymetry is assumed.
        :param gravity: Gravitational acceleration
        """
        self.b = b or (lambda x: np.zeros(x.shape))
        self.b_x = b_x or (lambda x: np.zeros(x.shape))
        self.gravity = gravity

    def hyperbolic_flux_term(self, u, xx, t):
        """
        SWE hyperbolic flux term:

        F(x) = [
          q
          q^2 / h + 1/2 * g * h^2
        ]

        :param u:
        :param xx:
        :param t:
        :return:
        """
        height, flow = u

        return np.array((
            flow,
            np.square(flow) / height + 0.5 * self.gravity * np.square(height)
        ))

    def hyperbolic_flux_jacobian(self, u, xx, t):
        """
        SWE hyperbolic flux Jacobian:

        F'(x) = [
          0                     1
          -q^2 / h^2 + g*h      2 * q / h
        ]

        :param u:
        :param xx:
        :param t:
        :return:
        """
        height, flow = u

        return np.array((
            (0,                                         1),
            (-(flow/height)**2 + self.gravity * height, 2 * flow/height)
        ))

    def source_term(self, u, xx, t):
        """
        SWE source term:

        S(x) = [
          0
          -g * h * b_x
        ]

        :param u:
        :param xx:
        :param t:
        :return:
        """
        height, flow = u

        return np.array((
            np.zeros(height.shape),
            -self.gravity * height * self.b_x(xx)
        ))

    def plot_animation(self, solution, frame_interval=20, animate_frame=None, initialize_plots=None):
        """
        Overrides base class implementation in order to add bathymetry to animations.

        :param solution:
        :param frame_interval:
        :param animate_frame:
        :param initialize_plots:
        :return:
        """
        return super().plot_animation(
            solution,
            frame_interval,
            animate_frame=self._animate_swe_frame,
            initialize_plots=self._initialize_swe_plots
        )

    def _initialize_swe_plots(self, axes, domain, soln, soln_shape):
        """
        Initializes plots for animation. Axis limits for each subplot are set to max/min values
        across entire time domain with respect to the subplot's solution variable.

        :param axes: Subplot axes.
        :param domain:
        :param soln: Solution indexed by time step
        :param soln_shape: Shape of solution variable
        :return:
        """
        plot_lines = []

        # Reshape solution into something a little more intuitive to work with. First dimension
        # is the time step index, second is the solution variable of interest, third dimension
        # is the cell index, and fourth is the node.
        #
        uu = soln.y.T.reshape(len(soln.t), *soln_shape)

        for i in range(0, self.solution_vars):
            # Plot solution variables
            #
            if i == 0:
                bb = self.b(domain)
                # Plot bathymetry
                #
                line, = axes[i].plot(domain.flatten(), uu[0, i].flatten() + bb.flatten())
                axes[i].plot(domain.flatten(), bb.flatten())
                plot_min = min(uu[:, i].min(), bb.min())
                plot_max = max(uu[:, i].max(), bb.max())
            else:
                # Plot flow
                #
                plot_min = (uu[:, i]).min()
                plot_max = (uu[:, i]).max()
                line, = axes[i].plot(domain.flatten(), (uu[0, i]).flatten())

                # Plot velocity
                #
                # plot_min = (uu[:, i]/uu[:, 0]).min()
                # plot_max = (uu[:, i]/uu[:, 0]).max()
                # line, = axes[i].plot(domain.flatten(), (uu[0, i]/uu[0, 0]).flatten())

            plot_lines.append(line)

            # Add a little extra padding to top and bottom of plots
            #
            padding = (0.1 * (plot_max - plot_min)) or 0.1
            axes[i].set_ylim(plot_min - padding, plot_max + padding)

        return plot_lines

    def _animate_swe_frame(self, n, axes, plot_lines, domain, soln, soln_shape):
        """
        Animate plot according to:

        https://matplotlib.org/stable/api/_as_gen/matplotlib.animation.FuncAnimation.html

        :param n: Frame index
        :param axes: Subplot axes.
        :param plot_lines: Plot lines to update. This should contain one line per solution.
        :param domain:
        :param soln: Solution indexed by time step
        :param soln_shape: Shape of solution variable

        :return:
        """
        u_i = soln.y[:, n].reshape(soln_shape)
        var_names = {
            0: 'h',
            1: 'hv'
        }

        # Plot each solution quantity
        #
        for k, line in enumerate(plot_lines):
            axes[k].set_title('${}; t = {:0.2f}$'.format(var_names[k], soln.t[n]))
            line.set_xdata(domain.flatten())
            if k == 0:
                bb = self.b(domain)
                # Add bathymetry plot
                #
                line.set_ydata(u_i[k].flatten() + bb.flatten())
            elif k == 1:
                # Plot flow
                #
                line.set_ydata((u_i[k]).flatten())

                # Plot velocity
                #
                # line.set_ydata((u_i[k]/u_i[0]).flatten())

        # This is the format required by the animate callback for plots.
        #
        return [(line,) for line in plot_lines]

    @staticmethod
    def surface_flux_fjordholm_etal(gravity=9.81):
        """
        Total energy conservative (mathematical entropy for shallow water equations). When the bottom topography
        is nonzero this should only be used as a surface flux otherwise the scheme will not be well-balanced.
        For well-balancedness in the volume flux use [`flux_wintermeyer_etal`](@ref).

        Details are available in Eq. (4.1) in the paper:
        - Ulrik S. Fjordholm, Siddhartha Mishr and Eitan Tadmor (2011)
          Well-balanced and energy stable schemes for the shallow water equations with discontinuous topography
          [DOI: 10.1016/j.jcp.2011.03.042](https://doi.org/10.1016/j.jcp.2011.03.042)

        :param gravity: Gravitational constant.
        """
        def _flux_function(u_l, u_r, xx, t, f, f_prime):
            # Unpack left and right state
            h_l, q_l = u_l
            h_r, q_r = u_r

            v_l = q_l / h_l
            v_r = q_r / h_r

            # Average each factor of products in flux
            h_avg = 0.5 * (h_l + h_r)
            v_avg = 0.5 * (v_l + v_r)
            p_avg = 0.25 * gravity * (np.square(h_l) + np.square(h_r))

            # Calculate fluxes depending on orientation
            f1 = h_avg * v_avg
            f2 = f1 * v_avg + p_avg

            return np.array((f1, f2))
        return _flux_function

    @staticmethod
    def volume_flux_wintermeyer_etal(gravity=9.81):
        """
        Total energy conservative (mathematical entropy for shallow water equations) split form.
        When the bottom topography is nonzero this scheme will be well-balanced when used as a `volume_flux`.
        The `surface_flux` should still use, e.g., [`flux_fjordholm_etal`](@ref).

        Further details are available in Theorem 1 of the paper:
        - Niklas Wintermeyer, Andrew R. Winters, Gregor J. Gassner and David A. Kopriva (2017)
          An entropy stable nodal discontinuous Galerkin method for the two dimensional
          shallow water equations on unstructured curvilinear meshes with discontinuous bathymetry
          [DOI: 10.1016/j.jcp.2017.03.036](https://doi.org/10.1016/j.jcp.2017.03.036)

        :param gravity: Gravitational constant.
        """
        def _flux_function(u_l, u_r, xx, t, f, f_prime):

            # Unpack left and right state
            #
            h_l, q_l = u_l
            h_r, q_r = u_r

            # Get the velocities on either side
            #
            v_l = q_l / h_l
            v_r = q_r / h_r

            # Average each factor of products in flux
            #
            v_avg = 0.5 * (v_l + v_r)
            p_avg = 0.5 * gravity * h_l * h_r

            # Calculate fluxes depending on orientation
            #
            f1 = 0.5 * (q_l + q_r)
            f2 = f1 * v_avg + p_avg

            return np.array((f1, f2))
        return _flux_function

    @staticmethod
    def bc_prescribed_inflow(q_in, gravity=9.81, surface_flux=fluxes.lax_friedrichs_flux):
        """
        Maintains a (possibly time-dependent) prescribed rate of flow at the inflow boundary by setting
        backward characteristics for upwinded values and cell values equal at the leftmost boundary.

        :param q_in: A function with time as its single parameter, which should return the rate of flow
                     at a particular time.
        :param gravity: Gravitational constant.
        :param surface_flux: Function used to compute numerical flux between adjacent cell states.
        :return:
        """
        def _flux_function(u_l, u_r, xx, t, f, f_prime):
            h_r, q_r = u_r

            # Backward characteristic
            #
            w_b = q_r / h_r - 2 * np.sqrt(gravity * h_r)

            # Evaluate prescribed bathymetry value at t
            #
            q = q_in(t)

            # We solve for the square root of h to avoid numerical issues with Newton's method
            #
            sqrt_h = optimize.newton(
                lambda hh: q / np.square(hh) - 2 * np.sqrt(gravity) * hh - w_b,
                x0=np.square(h_r),
                # fprime=lambda hh: -2 * q / hh ** 3 - 2 * np.sqrt(g),
                maxiter=500
            )

            h = np.square(sqrt_h)

            return surface_flux(np.array((h, q)), u_r, xx, t, f, f_prime)

        return _flux_function



import numpy as np

from scipy.integrate import OdeSolver

from scipy.integrate._ivp.rk import rk_step, RkDenseOutput
from scipy.integrate._ivp.common import (validate_max_step, validate_tol, select_initial_step,
                                         warn_extraneous, validate_first_step)


class RungeKutta(OdeSolver):
    """Base class for explicit Runge-Kutta methods."""
    C: np.ndarray = NotImplemented
    A: np.ndarray = NotImplemented
    B: np.ndarray = NotImplemented
    P: np.ndarray = NotImplemented
    order: int = NotImplemented
    error_estimator_order: int = NotImplemented
    n_stages: int = NotImplemented

    def __init__(self, fun, t0, y0, t_bound, max_step=np.inf,
                 rtol=1e-3, atol=1e-6, vectorized=False,
                 first_step=None, **extraneous):
        warn_extraneous(extraneous)
        super().__init__(fun, t0, y0, t_bound, vectorized,
                         support_complex=True)
        self.y_old = None
        self.max_step = validate_max_step(max_step)
        self.rtol, self.atol = validate_tol(rtol, atol, self.n)
        self.f = self.fun(self.t, self.y)
        if first_step is None:
            self.h_abs = select_initial_step(
                self.fun, self.t, self.y, self.f, self.direction,
                self.error_estimator_order, self.rtol, self.atol)
        else:
            self.h_abs = validate_first_step(first_step, t0, t_bound)
        self.K = np.empty((self.n_stages + 1, self.n), dtype=self.y.dtype)
        self.error_exponent = -1 / (self.error_estimator_order + 1)
        self.h_previous = None

    def _step_impl(self):
        t = self.t
        y = self.y

        max_step = self.max_step
        min_step = 10 * np.abs(np.nextafter(t, self.direction * np.inf) - t)

        if self.h_abs > max_step:
            h_abs = max_step
        elif self.h_abs < min_step:
            h_abs = min_step
        else:
            h_abs = self.h_abs

        if h_abs < min_step:
            return False, self.TOO_SMALL_STEP

        h = h_abs * self.direction
        t_new = t + h

        if self.direction * (t_new - self.t_bound) > 0:
            t_new = self.t_bound

        h = t_new - t
        h_abs = np.abs(h)

        y_new, f_new = rk_step(self.fun, t, y, self.f, h, self.A, self.B, self.C, self.K)

        self.h_previous = h
        self.y_old = y

        self.t = t_new
        self.y = y_new

        self.h_abs = h_abs
        self.f = f_new

        return True, None

    def _dense_output_impl(self):
        Q = self.K.T.dot(self.P)
        return RkDenseOutput(self.t_old, self.t, self.y_old, Q)


class SSPRK33(RungeKutta):
    order = 3
    error_estimator_order = 2
    n_stages = 3
    C = np.array([0, 1, 1/2])
    A = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [1/4, 1/4, 0]
    ])
    B = np.array([1/6, 1/6, 2/3])
    P = np.array([[1, -4 / 3, 5 / 9],
                  [0, 1, -2/3],
                  [0, 4/3, -8/9],
                  [0, -1, 1]])
